# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
"""Utilities used during AutoML training."""
import collections
import logging
import warnings
import ast
from typing import Any, cast, Callable, Dict, List, Optional, Set, Tuple, Union

import azureml.dataprep as dprep
import numpy as np
import pandas as pd
import scipy
from azureml._common._error_definition import AzureMLError
from azureml._common._error_definition.user_error import ArgumentBlankOrEmpty, ArgumentOutOfRange
from azureml._tracing._tracer_factory import get_tracer
from azureml.automl.core.automl_base_settings import AutoMLBaseSettings
from azureml.automl.core.config_utilities import _check_validation_config
from azureml.automl.core.constants import FeatureType as _FeatureType
from azureml.automl.core.constants import FeaturizationConfigMode
from azureml.automl.core.featurization.featurizationconfig import FeaturizationConfig
from azureml.automl.core.shared import constants
from azureml.automl.core.shared import utilities, logging_utilities
from azureml.automl.core.shared._diagnostics.automl_error_definitions import (
    AllFeaturesAreExcluded, AllTargetsNan, ConflictingValueForArguments, DatasetsFeatureCountMismatch,
    DatasetContainsInf, DataShapeMismatch, DuplicateColumns, ExperimentTimeoutForDataSize,
    FeatureUnsupportedForIncompatibleArguments, FeaturizationConfigColumnMissing, FeaturizationRequired,
    InconsistentColumnTypeInTrainValid, IndistinctLabelColumn, InputDatasetEmpty,
    InsufficientSampleSize, InvalidArgumentType, InvalidArgumentWithSupportedValues, InvalidInputDatatype,
    InvalidMetricForSingleValuedColumn, InvalidValuesInCVSplitColumn, InvalidValuesInData, MissingColumnsInData,
    MissingValidationConfig, NCrossValidationsExceedsTrainingRows, NonOverlappingColumnsInTrainValid,
    UnrecognizedFeatures, OnnxUnsupportedDatatype, OverlappingYminYmax, SampleCountMismatch, SampleWeightsUnsupported,
    UnhashableValueInColumn)
from azureml.automl.core.shared._diagnostics.contract import Contract
from azureml.automl.core.shared._diagnostics.validation import Validation
from azureml.automl.core.shared.constants import NumericalDtype, SupportedInputDatatypes, TelemetryConstants
from azureml.automl.core.shared.exceptions import (ConfigException,
                                                   DataException,
                                                   UserException,
                                                   ValidationException)
from azureml.automl.core.shared.reference_codes import ReferenceCodes
from azureml.automl.runtime import (
    dataprep_utilities, frequency_fixer, _time_series_training_utilities,
    _common_training_utilities)
from azureml.automl.runtime._automl_settings_utilities import rule_based_validation
from azureml.automl.runtime.column_purpose_detection import ColumnPurposeDetector
from azureml.automl.runtime.data_context import TransformedDataContext
from azureml.automl.runtime.featurizer.transformer.timeseries import forecasting_heuristic_utils
from azureml.automl.runtime.frequency_fixer import fix_data_set_regularity_may_be
from azureml.automl.runtime.shared import utilities as runtime_utilities
from azureml.automl.runtime.shared._cv_splits import _CVSplits
from azureml.automl.runtime.shared.cache_store import CacheStore
from azureml.automl.runtime.shared.datasets import SubsampleCacheStrategy, ClientDatasets, DatasetBase
from azureml.automl.runtime.shared.streaming_dataset import StreamingDataset, DatasetMetadataKeys
from azureml.automl.runtime.shared.types import CoreDataInputType, DataInputType, DataSingleColumnInputType
from azureml.automl.runtime.shared.utilities import _get_num_unique
from azureml.automl.runtime import short_grain_padding
from azureml.automl.runtime.streaming_data_context import StreamingTransformedDataContext
from azureml.dataprep import DataPrepException as DprepException
from azureml.dataprep.api.dataflow import DataflowValidationError
from sklearn.base import TransformerMixin
from sklearn.utils import validation as sk_validation

logger = logging.getLogger(__name__)
tracer = get_tracer(__name__)


class SmallDataSetLimit:
    """Constants for the small dataset limit."""

    WARNING_SIZE = 100
    MINIMAL_TRAIN_SIZE = 50
    MINIMAL_VALIDATION_SIZE = int(MINIMAL_TRAIN_SIZE / 10)


class LargeDatasetLimit:
    """Constants for limiting large datasets."""

    MAX_ROWS_TO_SUBSAMPLE = 100000

    # Number of rows to  use for doing validations on the data
    VALIDATION_SUBSAMPLE_SIZE = 5000


MASKED = '[Masked]'


def auto_block_models(input_data, automl_settings):
    """
    Add appropriate files to blocked_models automatically.

    :param input_data:
    :param automl_settings: The settings used for this current run.
    :return:
    """
    # Only enable auto-block if user didn't add any models to the allowed list
    if automl_settings.auto_blacklist and not automl_settings.whitelist_models:
        X = input_data['X']
        if scipy.sparse.issparse(X) or X.shape[0] > constants.MAX_SAMPLES_AUTOBLOCK:
            if automl_settings.blacklist_algos is None:
                automl_settings.blacklist_algos = \
                    constants.MAX_SAMPLES_AUTOBLOCKED_ALGOS
            else:
                for blacklist_algo in constants.MAX_SAMPLES_AUTOBLOCKED_ALGOS:
                    if blacklist_algo not in automl_settings.blacklist_algos:
                        automl_settings.blacklist_algos.append(blacklist_algo)
            automl_settings.blacklist_samples_reached = True
            automl_settings._validate_model_filter_lists()


def set_task_parameters(y: DataSingleColumnInputType, automl_settings: AutoMLBaseSettings) -> None:
    """
    Set this task's parameters based on some heuristics if they aren't provided.

    TODO: Move this code into AutoML settings or something. Client shouldn't have to think about this stuff.

    :param y: The list of possible output values.
    :param automl_settings: The settings used for this current run.
    :return: None.
    :raises: `ValidationException` in case of type errors. `DataException` in case of single valued target column
    """
    if automl_settings.enable_streaming:
        # Do not apply these heuristics if streaming is enabled.
        return

    Contract.assert_true(automl_settings.task_type in {constants.Tasks.CLASSIFICATION, constants.Tasks.REGRESSION},
                         "Invalid task type ({}) specified.".format(automl_settings.task_type))

    if automl_settings.task_type == constants.Tasks.CLASSIFICATION:
        #  Guess number of classes if the user did not explicitly provide it
        if not automl_settings.num_classes or not isinstance(
                automl_settings.num_classes, int):
            automl_settings.num_classes = _get_num_unique(pd.Series(y), ignore_na=True)
        return

    if automl_settings.task_type == constants.Tasks.REGRESSION:
        y_datatype = runtime_utilities._get_column_data_type_as_str(y)
        if not utilities._check_if_column_data_type_is_numerical(y_datatype):
            raise ValidationException._with_error(AzureMLError.create(
                InvalidInputDatatype,
                input_type=y_datatype,
                reference_code=ReferenceCodes._VALIDATE_TARGET_COLUMN_INCORRECT_TYPE,
                supported_types=", ".join(list(NumericalDtype.FULL_SET)),
                target='y'
            ))

        #  Guess min and max of y if the user did not explicitly provide it
        y_min = automl_settings.y_min
        y_max = automl_settings.y_max
        if not automl_settings.y_min:
            y_min = np.nanmin(y)
        if not automl_settings.y_max:
            y_max = np.nanmax(y)
        if y_max is not None and y_min is not None and y_max == y_min:
            raise DataException._with_error(AzureMLError.create(
                OverlappingYminYmax, target="y/y_min/y_max", value=str(y_min))
            )


def format_training_data(
        X=None, y=None, sample_weight=None, X_valid=None, y_valid=None, sample_weight_valid=None,
        cv_splits_indices=None, user_script=None,
        training_data=None, validation_data=None, label_column_name=None, weight_column_name=None,
        cv_split_column_names=None, is_adb_run=False, automl_settings=None, verifier=None):
    """
    Create a dictionary with training and validation data from all supported input formats.

    :param X: Training features.
    :type X: pandas.DataFrame or numpy.ndarray or azureml.dataprep.Dataflow
    :param y: Training labels.
    :type y: pandas.DataFrame or numpy.ndarray or azureml.dataprep.Dataflow
    :param sample_weight: Sample weights for training data.
    :type sample_weight: pandas.DataFrame pr numpy.ndarray or azureml.dataprep.Dataflow
    :param X_valid: validation features.
    :type X_valid: pandas.DataFrame or numpy.ndarray or azureml.dataprep.Dataflow
    :param y_valid: validation labels.
    :type y_valid: pandas.DataFrame or numpy.ndarray or azureml.dataprep.Dataflow
    :param sample_weight_valid: validation set sample weights.
    :type sample_weight_valid: pandas.DataFrame or numpy.ndarray or azureml.dataprep.Dataflow
    :param cv_splits_indices:
        Indices where to split training data for cross validation.
        Each row is a separate cross fold and within each crossfold, provide 2 arrays,
        the first with the indices for samples to use for training data and the second
        with the indices to use for validation data. i.e [[t1, v1], [t2, v2], ...]
        where t1 is the training indices for the first cross fold and v1 is the validation
        indices for the first cross fold.
    :type cv_splits_indices: numpy.ndarray
    :param training_data: The training data to be used within the experiment.
    :type training_data: pandas.DataFrame or azureml.core.Dataset
        or azureml.data.dataset_definition.DatasetDefinition or azureml.data.TabularDataset
    :param validation_data: The validation data to be used within the experiment.
    :type validation_data: pandas.DataFrame or azureml.core.Dataset
        or azureml.data.dataset_definition.DatasetDefinition or azureml.data.TabularDataset
    :param label_column_name: The name of the label column.
    :type label_column_name: str
    :param weight_column_name: The name of the sample weight column.
    :type weight_column_name: str
    :param cv_split_column_names: List of names for columns that contain custom cross validation split.
    :type cv_split_column_names: list(str)
    :param user_script: File path to script containing get_data()
    :param is_adb_run: True if this is being called from an ADB/local experiment
    :param automl_settings: automl settings
    :param verifier: Verifier Manager instance.
    :type verifier: Optional[VerifierManager]
    :return:
    """
    data_dict = None
    x_raw_column_names = None
    reset_index_msg = ("Reset dataframe index to avoid potential conflicts "
                       "if time/time_series_id_column_names column is part of an index.")

    with tracer.start_as_current_span(
            TelemetryConstants.DataFetch,
            user_facing_name=TelemetryConstants.DataFetch
    ):
        if X is None and y is None and training_data is None:
            if data_dict is None:
                data_dict = _extract_user_data(user_script)
            X = data_dict.get('X')
            y = data_dict.get('y')
            sample_weight = data_dict.get('sample_weight')
            X_valid = data_dict.get('X_valid')
            y_valid = data_dict.get('y_valid')
            sample_weight_valid = data_dict.get('sample_weight_valid')
            cv_splits_indices = data_dict.get("cv_splits_indices")
            x_raw_column_names = data_dict.get("x_raw_column_names")
        elif training_data is not None and label_column_name is not None:
            if isinstance(training_data, pd.DataFrame):
                x_raw_column_names = training_data.columns.values
                X, y, sample_weight, cv_splits_indices = _extract_data_from_combined_dataframe(
                    training_data=training_data, label_column_name=label_column_name,
                    sample_weight_column_name=weight_column_name, cv_split_column_names=cv_split_column_names
                )

                if validation_data is not None:
                    X_valid, y_valid, sample_weight_valid, _ = _extract_data_from_combined_dataframe(
                        training_data=validation_data, label_column_name=label_column_name,
                        sample_weight_column_name=weight_column_name
                    )
            elif isinstance(training_data, dprep.Dataflow):
                X, y, sample_weight, cv_splits_indices = _extract_data_from_combined_dataflow(
                    training_data=training_data, label_column_name=label_column_name,
                    sample_weight_column_name=weight_column_name, cv_split_column_names=cv_split_column_names
                )

                if validation_data is not None:
                    X_valid, y_valid, sample_weight_valid, _ = _extract_data_from_combined_dataflow(
                        training_data=validation_data, label_column_name=label_column_name,
                        sample_weight_column_name=weight_column_name
                    )
                x_raw_column_names = X.head(1).columns.values

    # Get the raw column names
    if isinstance(X, pd.DataFrame):
        # reset index in case a customer's df contains index column(s)
        X.reset_index(inplace=True, drop=True)
        forecasting_heuristic_utils._log_warn_maybe(reset_index_msg)
        # Cache the raw column names if available
        x_raw_column_names = X.columns.values
    else:
        as_numpy = True
        if is_adb_run:
            # Hack to make sure we get a pandas DF and not a numpy array in ADB
            # The two retrieval functions should be rationalized in future releases
            as_numpy = False
        X = dataprep_utilities.materialize_dataflow(X, as_numpy=as_numpy)
        if X_valid is not None:
            X_valid = dataprep_utilities.materialize_dataflow(X_valid, as_numpy=as_numpy)

        y = dataprep_utilities.materialize_dataflow(y, as_numpy=True)
        if y_valid is not None:
            y_valid = dataprep_utilities.materialize_dataflow(y_valid, as_numpy=True)

        if sample_weight is not None:
            sample_weight = dataprep_utilities.materialize_dataflow(sample_weight, as_numpy=True)
        if sample_weight_valid is not None:
            sample_weight_valid = dataprep_utilities.materialize_dataflow(sample_weight_valid, as_numpy=True)

        if cv_splits_indices is not None and cv_split_column_names is None:
            # cv_splits_indices NOT derived from cv_split_column_names so it still needs to resolve
            cv_splits_indices = dataprep_utilities.resolve_cv_splits_indices(cv_splits_indices)

        if isinstance(X, pd.DataFrame):
            # reset index in case a customer's df contains index column(s)
            X.reset_index(inplace=True, drop=True)
            forecasting_heuristic_utils._log_warn_maybe(reset_index_msg)
            x_raw_column_names = X.columns.values

    # reset index in case a customer's df contains index column(s)
    if X_valid is not None and isinstance(X_valid, pd.DataFrame):
        X_valid.reset_index(inplace=True, drop=True)
        forecasting_heuristic_utils._log_warn_maybe(reset_index_msg)

    if automl_settings is None or \
            (automl_settings.featurization == FeaturizationConfigMode.Off and not automl_settings.is_timeseries):
        if isinstance(X, pd.DataFrame):
            X = X.values
        if isinstance(X_valid, pd.DataFrame):
            X_valid = X_valid.values
    y = _convert_to_numpy_maybe(y, 'y')
    y_valid = _convert_to_numpy_maybe(y_valid, 'y_valid')
    if isinstance(sample_weight, pd.DataFrame):
        sample_weight = sample_weight.values
    if isinstance(sample_weight_valid, pd.DataFrame):
        sample_weight_valid = sample_weight_valid.values
    if automl_settings is not None:
        # When data were read try to fix the frequency.
        if automl_settings.is_timeseries and X_valid is None:
            _check_dimensions(X, y, None, None, None, None)
            fixed_data_set = fix_data_set_regularity_may_be(
                X, y,
                automl_settings.time_column_name,
                automl_settings.grain_column_names,
                frequency_fixer.str_to_offset_safe(automl_settings.freq,
                                                   ReferenceCodes._TRAINING_UTILITIES_CHECK_FREQ_FIX))
            X = fixed_data_set.data_x
            y = fixed_data_set.data_y
            failed = fixed_data_set.is_failed
            was_corrected = fixed_data_set.is_modified
            freq = fixed_data_set.freq
            if verifier:
                verifier.update_data_verifier_frequency_inference(failed, was_corrected)
            # Pad the short grains if needed (the short_series_handing_config_value is
            # checked by pad_short_grains_or_raise).
            X, y = short_grain_padding.pad_short_grains_or_raise(
                X, y, freq, automl_settings, ReferenceCodes._TS_ONE_VALUE_PER_GRAIN_TSUTIL, verifier)

        # auto cv needs the input X and y have same rows
        _check_dimensions(
            X=X, y=y, X_valid=X_valid, y_valid=y_valid,
            sample_weight=sample_weight, sample_weight_valid=sample_weight_valid)

        X, y, sample_weight, X_valid, y_valid, sample_weight_valid = rule_based_validation(
            automl_settings=automl_settings,
            X=X,
            y=y,
            sample_weight=sample_weight,
            X_valid=X_valid,
            y_valid=y_valid,
            sample_weight_valid=sample_weight_valid,
            cv_splits_indices=cv_splits_indices,
            verifier=verifier
        )

    data_dict = {
        'X': X,
        'y': y,
        'X_valid': X_valid,
        'y_valid': y_valid,
        'cv_splits_indices': cv_splits_indices,
        'x_raw_column_names': x_raw_column_names,
        'sample_weight': sample_weight,
        'sample_weight_valid': sample_weight_valid}
    return data_dict


def _convert_to_numpy_maybe(
        y: Optional[Union[np.ndarray, pd.DataFrame, pd.Series]],
        ds_name: str) -> Optional[np.ndarray]:
    """
    Try to convert y to numpy array.

    If y can not be converted to np.ndarray or has wrong shape the DataException is raised.
    :param y: The data set to be converted.
    :param ds_name: The name of the data set to convert.
    :raises: DataException
    """
    if y is None:
        return y
    if isinstance(y, pd.DataFrame):
        _check_y_shape(y, 'y')
        return cast(np.ndarray, y[y.columns[0]].values)
    if isinstance(y, pd.Series):
        return cast(np.ndarray, y.values)
    return y


def _check_y_shape(y: pd.DataFrame, ds_name: str) -> None:
    """
    Check if y data frame has only one column.

    :param y: The y dataframe.
    :param name: The name of a data set.
    :raises: DataException
    """
    if y.shape[1] > 1:
        raise DataException._with_error(AzureMLError.create(
            DataShapeMismatch, target=ds_name, data_object_name=ds_name, expected_dim=1, actual_dim=y.shape[1],
            reference_code=ReferenceCodes._Y_SHAPE_MISMATCH)
        )


def validate_training_data(X: DataInputType,
                           y: DataInputType,
                           X_valid: Optional[DataInputType],
                           y_valid: Optional[DataInputType],
                           sample_weight: Optional[DataInputType],
                           sample_weight_valid: Optional[DataInputType],
                           cv_splits_indices: Optional[np.ndarray],
                           automl_settings: AutoMLBaseSettings,
                           check_sparse: bool = False,
                           x_raw_column_names: Optional[np.ndarray] = None) -> None:
    """
    Validate that training data and parameters have been correctly provided.

    :param X:
    :param y:
    :param X_valid:
    :param y_valid:
    :param sample_weight:
    :param sample_weight_valid:
    :param cv_splits_indices:
    :param automl_settings:
    :param check_sparse:
    :param x_raw_column_names: Raw column names as list of str.
    """
    # if using incremental learning, validate and subsample data inputs. subsampling the input Dataflows
    # to numpy arrays this allows the validation flow (which can handle numpy arrays but not
    # Dataflows directly at the moment) to proceed.

    check_x_y(X, y, automl_settings, x_valid=X_valid, y_valid=y_valid,
              check_sparse=check_sparse)

    # Ensure at least one form of validation is specified
    if not ((X_valid is not None) or automl_settings.n_cross_validations or
            (cv_splits_indices is not None) or automl_settings.validation_size):
        raise DataException._with_error(AzureMLError.create(MissingValidationConfig, target=", ".join(
            ["validation_data", "validation_size", "n_cross_validations"])))

    # validate sample weights if not None
    if sample_weight is not None:
        check_sample_weight(X, sample_weight, automl_settings)

    if sample_weight_valid is not None:
        check_sample_weight(X_valid, sample_weight_valid, automl_settings, True)

    _check_dimensions(
        X=X, y=y, X_valid=X_valid, y_valid=y_valid,
        sample_weight=sample_weight, sample_weight_valid=sample_weight_valid)

    _check_validation_config(X_valid=X_valid,
                             y_valid=y_valid,
                             sample_weight=sample_weight,
                             sample_weight_valid=sample_weight_valid,
                             cv_splits_indices=cv_splits_indices,
                             n_cross_validations=automl_settings.n_cross_validations,
                             validation_size=automl_settings.validation_size)

    if automl_settings.n_cross_validations is not None:
        if y.shape[0] < automl_settings.n_cross_validations:
            raise ConfigException._with_error(
                AzureMLError.create(
                    NCrossValidationsExceedsTrainingRows, target="n_cross_validations",
                    training_rows=y.shape[0], n_cross_validations=automl_settings.n_cross_validations
                )
            )

    # streaming uses subsampling + streaming does not use pandas, so the following isn't applicable - hence skip
    if automl_settings.featurization != FeaturizationConfigMode.Off and not automl_settings.enable_streaming:
        _check_data_can_be_preprocessed(X, X_valid, x_raw_column_names)

    metrics_checks(X, y, automl_settings, X_valid, y_valid)

    _validate_exp_timeout_for_data(X, automl_settings)
    _validate_featurization_config(automl_settings.featurization, x_raw_column_names)

    _validate_all_column_not_ignored(X, x_raw_column_names, automl_settings)


def validate_streaming_data(
        training_data: dprep.Dataflow,
        automl_settings: AutoMLBaseSettings,
        validation_data: Optional[dprep.Dataflow] = None) -> None:
    """
    If streaming is enabled, we do a best effort based validation of the inputs (due to potentially large data
    sizes), in order to not run out of memory.

    Note: We should refactor this method to have a strong contract in terms of what's expected as inputs. Currently,
    validation_data is optional, but under most (known) circumstances, this is always passed down from the caller.
    This is losely enforced as of now, i.e. it can be None, given the user had specified a validation_size. But for
    streaming, we always do a split of training data into validation data, either based on heuristics, or user given
    validation_size. This should be strongly enforced after subsequent refactorings.

    :param training_data: Input training data provided by the user
    :param automl_settings: AutoML configuration
    :param validation_data: Optional Validation data that the user can provide
    :return: None, if the validation succeeds
    """
    # Expecting streaming cases to only work on Dataflows
    Contract.assert_type(training_data, "training_data", expected_types=dprep.Dataflow,
                         reference_code=ReferenceCodes._STREAMING_INVALID_TRAINING_DATATYPE)
    if validation_data is not None:
        Contract.assert_type(validation_data, "validation_data", expected_types=dprep.Dataflow,
                             reference_code=ReferenceCodes._STREAMING_INVALID_VALIDATION_DATATYPE)

    label_column_name = automl_settings.label_column_name
    weight_column_name = automl_settings.weight_column_name

    columns_list = []
    y = None

    sample_weight = None  # type: Optional[DataInputType]
    sample_weight_valid = None  # type: Optional[DataInputType]

    if label_column_name is not None:
        columns_list.append(label_column_name)
        y = training_data.keep_columns(label_column_name)
    if weight_column_name is not None:
        columns_list.append(weight_column_name)
        sample_weight = training_data.keep_columns(weight_column_name)
        if automl_settings.validation_size == 0 and validation_data:
            sample_weight_valid = validation_data.keep_columns(weight_column_name)

    X = training_data.drop_columns(columns_list)

    X_valid = None  # type: Optional[DataInputType]
    y_valid = None  # type: Optional[DataInputType]
    # We don't need to set X_valid / y_valid in case the user specified validation_size, since in this case the
    # validation data is just a part of the overall training data.
    if automl_settings.validation_size == 0 and validation_data:
        X_valid = validation_data.drop_columns(columns_list)
        y_valid = validation_data.keep_columns(label_column_name) \
            if label_column_name is not None else None

    X, y, X_valid, y_valid, sample_weight, sample_weight_valid = _incremental_learning_validate_and_subsample_inputs(
        X, y, X_valid, y_valid, sample_weight, sample_weight_valid)

    validate_training_data(X=X, y=y, X_valid=X_valid, y_valid=y_valid, sample_weight=sample_weight,
                           sample_weight_valid=sample_weight_valid, cv_splits_indices=None,
                           automl_settings=automl_settings)


def _incremental_learning_validate_and_subsample_inputs(
    X: dprep.Dataflow,
    y: dprep.Dataflow,
    X_valid: Optional[dprep.Dataflow] = None,
    y_valid: Optional[dprep.Dataflow] = None,
    sample_weight: Optional[dprep.Dataflow] = None,
    sample_weight_valid: Optional[dprep.Dataflow] = None
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    If using incremental learning, validate and subsample data inputs.
    Subsampling the input Dataflows to numpy arrays this allows the validation flow (which
    can handle numpy arrays but not Dataflows directly at the moment) to proceed.

    :param X:
    :param y:
    :param X_valid:
    :param y_valid:
    :param sample_weight:
    :param sample_weight_valid:
    """
    try:
        # validate that y is only a single column
        y_column_count = len(y.head(1).columns)
        if y_column_count != 1:
            raise DataException._with_error(AzureMLError.create(
                DataShapeMismatch, target="y", data_object_name="y", expected_dim=1, actual_dim=y_column_count)
            )

        # validate that column names are unique between X and y
        # (this is required, b/c we append the columns of X and y together to get a single merged Dataflow
        # that nimbus learners require as input. if X and y have shared column names, this merge throws an error)
        X_column_names = X.head(1).columns.tolist()
        y_column_name = y.head(1).columns[0]
        if y_column_name in X_column_names:
            raise DataException._with_error(AzureMLError.create(
                IndistinctLabelColumn, target='label_column_name', label_column_name=y_column_name,
                reference_code=ReferenceCodes._INDISTINCT_LABEL_COLUMN_IN_DATAFLOW)
            )

        # validate that the Datatypes are the same for X and y (for training, validation and sample weight data)
        for train, valid in [(X, X_valid), (y, y_valid), (sample_weight, sample_weight_valid)]:
            _check_datatypes(train, valid)

        # generate subsampled numpy arrays
        X = X.take(LargeDatasetLimit.VALIDATION_SUBSAMPLE_SIZE).to_pandas_dataframe().values
        y = y.take(LargeDatasetLimit.VALIDATION_SUBSAMPLE_SIZE).to_pandas_dataframe().iloc[:, 0].values
        if X_valid is not None:
            X_valid = X_valid.take(LargeDatasetLimit.VALIDATION_SUBSAMPLE_SIZE).to_pandas_dataframe().values
        if y_valid is not None:
            y_valid = y_valid.take(LargeDatasetLimit.VALIDATION_SUBSAMPLE_SIZE).to_pandas_dataframe().iloc[:, 0].values
        if sample_weight is not None:
            sample_weight = (sample_weight.take(LargeDatasetLimit.VALIDATION_SUBSAMPLE_SIZE).
                             to_pandas_dataframe().iloc[:, 0].values)
        if sample_weight_valid is not None:
            sample_weight_valid = (sample_weight_valid.take(LargeDatasetLimit.VALIDATION_SUBSAMPLE_SIZE).
                                   to_pandas_dataframe().iloc[:, 0].values)
    except DprepException as e:
        logging_utilities.log_traceback(e, logger)
        dataprep_utilities.dataprep_error_handler(e)

    return \
        cast(np.ndarray, X),\
        cast(np.ndarray, y),\
        cast(np.ndarray, X_valid),\
        cast(np.ndarray, y_valid),\
        cast(np.ndarray, sample_weight),\
        cast(np.ndarray, sample_weight_valid)


def _check_datatypes(training_data: dprep.Dataflow, validation_data: dprep.Dataflow) -> None:
    if training_data is not None and validation_data is not None:
        train_dtypes = (training_data.take(LargeDatasetLimit.VALIDATION_SUBSAMPLE_SIZE).dtypes.items())
        valid_dtypes = (validation_data.take(LargeDatasetLimit.VALIDATION_SUBSAMPLE_SIZE).dtypes.items())

        for (train_col, train_datatype), (valid_col, valid_datatype) in zip(train_dtypes, valid_dtypes):
            if train_datatype != valid_datatype and train_col == valid_col:
                raise ValidationException._with_error(AzureMLError.create(
                    InconsistentColumnTypeInTrainValid, target=train_col, column_name=train_col,
                    train_dtype=train_datatype, validation_dtype=valid_datatype)
                )


def validate_training_data_dict(data_dict: Dict[str, Any],
                                automl_settings: AutoMLBaseSettings,
                                check_sparse: bool = False) -> None:
    """
    Validate that training data and parameters have been correctly provided.

    :param data_dict: Key-Value pair containing information about user provided data
    :param automl_settings: AutoML configuration
    :param check_sparse:
    :return:
    """
    logger.info("Start data validation.")
    with tracer.start_as_current_span(
        TelemetryConstants.DataValidation,
        user_facing_name=TelemetryConstants.DataValidation
    ):
        if automl_settings.enable_streaming:
            if data_dict.get('cv_splits_indices') is not None:
                raise ConfigException._with_error(
                    AzureMLError.create(
                        FeatureUnsupportedForIncompatibleArguments, target="cv_splits_indices",
                        feature_name='StreamingTrain', arguments="cv_splits_indices"
                    )
                )

            training_data = data_dict.get('training_data')
            validation_data = data_dict.get('validation_data')

            validate_streaming_data(training_data, automl_settings, validation_data)
            logger.info("Streaming was enabled, successfully validated data.")
            return

        X = data_dict.get('X', None)
        y = data_dict.get('y', None)
        sample_weight = data_dict.get('sample_weight', None)
        X_valid = data_dict.get('X_valid', None)
        y_valid = data_dict.get('y_valid', None)
        sample_weight_valid = data_dict.get('sample_weight_valid', None)
        cv_splits_indices = data_dict.get('cv_splits_indices', None)
        x_raw_column_names = data_dict.get('x_raw_column_names', None)
        validate_training_data(X, y, X_valid, y_valid, sample_weight, sample_weight_valid, cv_splits_indices,
                               automl_settings, check_sparse=check_sparse, x_raw_column_names=x_raw_column_names)
        if automl_settings.is_timeseries:
            _time_series_training_utilities.validate_timeseries_training_data(
                automl_settings, X, y, X_valid, y_valid,
                sample_weight, sample_weight_valid, cv_splits_indices,
                x_raw_column_names)


def metrics_checks(x: DataInputType,
                   y: DataInputType,
                   automl_settings: AutoMLBaseSettings,
                   x_valid: Optional[DataInputType] = None,
                   y_valid: Optional[DataInputType] = None) -> None:
    """
    Validate input data for metrics.

    :param x: input data. dataframe/ array/ sparse matrix
    :param y: input labels. dataframe/series/array
    :param automl_settings: automl settings
    :raise: DataException if data is not suitable for metrics calculations
    :return:
    """
    if y_valid is not None:
        # since we subsample the data for streaming following calculations may be inaccurate - hence skip them
        if automl_settings.task_type == constants.Tasks.CLASSIFICATION and not automl_settings.enable_streaming:
            primary_metric = automl_settings.primary_metric
            if primary_metric == constants.Metric.AUCWeighted:
                in_valid = set(y_valid[~pd.isnull(y_valid)])
                if len(in_valid) == 1:
                    remaining_metrics = utilities.get_primary_metrics(constants.Tasks.CLASSIFICATION).copy()
                    remaining_metrics.remove(primary_metric)
                    raise DataException._with_error(AzureMLError.create(
                        InvalidMetricForSingleValuedColumn, target="y_valid", data_object_name="y_valid",
                        valid_primary_metrics=remaining_metrics)
                    )


def check_x_y(x: DataInputType,
              y: DataInputType,
              automl_settings: AutoMLBaseSettings,
              x_valid: Optional[DataInputType] = None,
              y_valid: Optional[DataInputType] = None,
              check_sparse: bool = False) -> None:
    """
    Validate input data.

    :param x: input data. dataframe/ array/ sparse matrix
    :param y: input labels. dataframe/series/array
    :param automl_settings: automl settings
    :raise: DataException if data does not conform to accepted types and shapes
    :return:
    """

    logger.info('Checking X and y.')

    Validation.validate_value(x, "X")
    Validation.validate_value(y, "y")

    is_timeseries = automl_settings.is_timeseries
    is_featurization_enabled = automl_settings.featurization != FeaturizationConfigMode.Off

    if isinstance(x, pd.DataFrame):
        if len(x.columns) != len(set(x.columns)):
            column_list = list(x.columns)
            duplicate_columns = [col for col, count in collections.Counter(column_list).items() if count > 1]
            raise DataException._with_error(AzureMLError.create(
                DuplicateColumns, target='X', data_object_name='X', duplicate_columns=", ".join(duplicate_columns))
            )
        if x_valid is not None and isinstance(x_valid, pd.DataFrame):
            if len(x.columns.intersection(x_valid.columns)) != len(x.columns):
                x_column_list = list(x.columns)
                x_valid_column_list = list(x_valid.columns)
                missing_columns = set([col for col in x_column_list if x_valid_column_list.count(col) == 0])
                raise DataException._with_error(AzureMLError.create(
                    NonOverlappingColumnsInTrainValid, target='X', missing_columns=", ".join(missing_columns))
                )

    if x_valid is not None:
        if len(x.shape) > 1:
            if len(x_valid.shape) > 1 and x.shape[1] != x_valid.shape[1]:
                raise DataException._with_error(AzureMLError.create(
                    DatasetsFeatureCountMismatch, target="X/X_Valid",
                    first_dataset_name="X", first_dataset_shape=x.shape[1],
                    second_dataset_name="X_valid", second_dataset_shape=x_valid.shape[1])
                )
            elif len(x_valid.shape) == 1 and x.shape[1] != 1:
                raise DataException._with_error(
                    AzureMLError.create(
                        DatasetsFeatureCountMismatch, target="X/X_Valid", first_dataset_name="X",
                        first_dataset_shape=x.shape[1], second_dataset_name="X_valid", second_dataset_shape=1
                    )
                )
        elif len(x_valid.shape) > 1 and x_valid.shape[1] != 1:
            raise DataException._with_error(
                AzureMLError.create(
                    DatasetsFeatureCountMismatch, target="X/X_Valid", first_dataset_name="X",
                    first_dataset_shape=x.shape[1], second_dataset_name="X_valid",
                    second_dataset_shape=x_valid.shape[1]
                )
            )

    if not (((is_featurization_enabled or is_timeseries) and isinstance(x, pd.DataFrame)) or
            isinstance(x, np.ndarray) or scipy.sparse.issparse(x)):
        raise ValidationException._with_error(
            AzureMLError.create(InvalidArgumentType, target="X", argument="X", actual_type=type(x),
                                expected_types="pandas.DataFrame, numpy.ndarray, scipy.sparse")
        )

    if (check_sparse and scipy.sparse.issparse(x) and
        (automl_settings.enable_onnx_compatible_models is True or
         automl_settings.enable_onnx_compatible_models == "True")):
        raise ValidationException._with_error(
            AzureMLError.create(OnnxUnsupportedDatatype, supported_datatypes="pandas.DataFrame, numpy.ndarray")
        )

    Validation.validate_type(y, "y", np.ndarray)

    if len(y.shape) > 2 or (len(y.shape) == 2 and y.shape[1] != 1):
        raise DataException._with_error(AzureMLError.create(
            DataShapeMismatch, target="y", data_object_name="y", expected_dim=1, actual_dim=len(y.shape))
        )

    if y is not None:
        if len(runtime_utilities._get_indices_missing_labels_output_column(y)) == y.shape[0]:
            raise DataException._with_error(AzureMLError.create(AllTargetsNan, target="y"))

    if y_valid is not None:
        if len(runtime_utilities._get_indices_missing_labels_output_column(y_valid)) == y_valid.shape[0]:
            raise DataException._with_error(AzureMLError.create(AllTargetsNan, target="y_valid"))

    if automl_settings.task_type == constants.Tasks.REGRESSION:
        if not utilities._check_if_column_data_type_is_numerical(
                runtime_utilities._get_column_data_type_as_str(y)):
            raise ValidationException._with_error(AzureMLError.create(
                InvalidArgumentType, target="y", argument="y", actual_type=type(x),
                expected_types=NumericalDtype.FULL_SET)
            )

    # If text data is not being preprocessed or featurized, then raise an error
    if not is_featurization_enabled and not is_timeseries:
        types_requiring_featurization = "{}, {} or {}".format(
            _FeatureType.DateTime.lower(), _FeatureType.Categorical.lower(), _FeatureType.Text.lower()
        )

        # counter to keep track of how many numerical columns are marked as Ignore or AllNan type
        numeric_column_drop_set_counter = 0
        if isinstance(x, pd.DataFrame):
            for column in x.columns:
                if not utilities._check_if_column_data_type_is_numerical(
                        runtime_utilities._get_column_data_type_as_str(x[column].values)):
                    raise ValidationException._with_error(
                        AzureMLError.create(FeaturizationRequired, feature_types=types_requiring_featurization)
                    )
                elif _is_numeric_x_part_of_drop_set(x[column]):
                    numeric_column_drop_set_counter += 1
            if numeric_column_drop_set_counter == len(x.columns):
                raise DataException._with_error(AzureMLError.create(AllFeaturesAreExcluded, target="X"))
        elif isinstance(x, np.ndarray):
            if len(x.shape) == 1:
                if not utilities._check_if_column_data_type_is_numerical(
                        runtime_utilities._get_column_data_type_as_str(x)):
                    raise ValidationException._with_error(
                        AzureMLError.create(FeaturizationRequired, feature_types=types_requiring_featurization)
                    )
                elif _is_numeric_x_part_of_drop_set(pd.Series(x)):
                    raise DataException._with_error(AzureMLError.create(AllFeaturesAreExcluded, target="X"))
            else:
                for index in range(x.shape[1]):
                    if not utilities._check_if_column_data_type_is_numerical(
                            runtime_utilities._get_column_data_type_as_str(x[:, index])):
                        raise ValidationException._with_error(
                            AzureMLError.create(FeaturizationRequired, feature_types=types_requiring_featurization)
                        )
                    elif _is_numeric_x_part_of_drop_set(pd.Series(x[:, index])):
                        numeric_column_drop_set_counter += 1
                if numeric_column_drop_set_counter == x.shape[1]:
                    raise DataException._with_error(AzureMLError.create(AllFeaturesAreExcluded, target="X"))

    # since we subsample the data for streaming following calculations may be inaccurate - hence skip them
    if not automl_settings.enable_streaming:
        _common_training_utilities._column_value_number_check(
            y, automl_settings.task_type)

    # not check x Nan if featurization is enabled.
    check_x_nan = not is_featurization_enabled
    # not check NaN in y data as we will automatically remove these data in the data_transformation.py.
    check_y_nan = False
    # always check x contains inf or not.
    check_x_inf = True
    # check y contains inf data raise errors and only in regression.
    check_y_inf = automl_settings.task_type != constants.Tasks.CLASSIFICATION
    _check_data_nan_inf(
        x, input_data_name="X", check_nan=check_x_nan, check_inf=check_x_inf)
    _check_data_nan_inf(y, input_data_name="y", check_nan=check_y_nan, check_inf=check_y_inf)
    # forecasting task has its own check for minimal data check here for regression and classification.
    if not automl_settings.is_timeseries:
        _check_data_minimal_size(x, x_valid, automl_settings)
    if x_valid is not None:
        _check_data_nan_inf(
            x_valid, input_data_name="X_valid", check_nan=check_x_nan, check_inf=check_x_inf)
    if y_valid is not None:
        _check_data_nan_inf(
            y_valid, input_data_name="y_valid", check_nan=check_y_nan, check_inf=check_y_inf)


def _is_numeric_x_part_of_drop_set(x: pd.Series) -> bool:
    # Numerical column with feature type of Ignore or AllNan does not go through featurization.
    # If dataset contains all numerical with Ignore or AllNan, then we should alert the user.
    non_na_raw_column = x.dropna()
    return not non_na_raw_column.shape[0] or non_na_raw_column.unique().shape[0] == 1


def _check_data_minimal_size(X: DataInputType, X_valid: DataInputType, automl_settings: AutoMLBaseSettings) -> None:
    """Check if the data is larger than minimum size."""

    if X.shape[0] < SmallDataSetLimit.MINIMAL_TRAIN_SIZE:
        raise DataException._with_error(AzureMLError.create(
            InsufficientSampleSize, target="X", data_object_name="X", sample_count=X.shape[0],
            minimum_count=SmallDataSetLimit.MINIMAL_TRAIN_SIZE)
        )
    if X_valid is not None and X_valid.shape[0] < SmallDataSetLimit.MINIMAL_VALIDATION_SIZE:
        raise DataException._with_error(AzureMLError.create(
            InsufficientSampleSize, target="X_valid", data_object_name="X_valid", sample_count=X_valid.shape[0],
            minimum_count=SmallDataSetLimit.MINIMAL_VALIDATION_SIZE)
        )
    if X.shape[0] < SmallDataSetLimit.WARNING_SIZE:
        warnings.warn(
            "The input data X has {} data points which is less than the recommended "
            "minimum data size {}. Please consider adding more data points to ensure better model accuracy.".
            format(X.shape[0], SmallDataSetLimit.WARNING_SIZE)
        )


def _check_data_nan_inf(data: DataInputType,
                        input_data_name: str,
                        check_nan: bool,
                        check_inf: bool = True) -> None:
    """Check if data contains nan or inf. If contains NaN, give out warning. If contains inf, raise exception."""
    if isinstance(data, pd.DataFrame):
        data_type = data.dtypes.dtype
    else:
        data_type = data.dtype
    is_integer_data = data_type.char in np.typecodes['AllInteger']
    n_top_indices = 20
    try:
        _check_all_finite(data, input_data_name)
        if check_nan and is_integer_data:
            # if the data is all integer, we will have a nan check beyond what sklearn does.
            input_data = data.data if scipy.sparse.issparse(data) else data
            if any(np.isnan(input_data)):
                raise ValueError
    except ValueError as ve:
        # looking for nan and inf for the data. If the data contains other type, it will used in other checks.
        if data_type.char in np.typecodes['AllFloat'] or (check_nan and is_integer_data):
            if check_nan:
                nan_indices = _get_data_indices_by_mask_function(data, np.isnan)
                if nan_indices.shape[0] > 0:
                    print(
                        "WARNING: The following coordinates{} [{}] contains {} NaN(np.NaN) data in {}. "
                        "Please consider dropping these rows.".
                        format(_construct_coord_indices_str(nan_indices, n_top_indices),
                               "" if nan_indices.shape[0] < n_top_indices else "(first detected in each column)",
                               nan_indices.shape[0],
                               input_data_name)
                    )
            if check_inf:
                inf_indices = _get_data_indices_by_mask_function(data, np.isinf)
                if inf_indices.shape[0] > 0:
                    raise DataException._with_error(
                        AzureMLError.create(
                            DatasetContainsInf, target=input_data_name, data_object_name=input_data_name,
                            inf_count=inf_indices.shape[0],
                            coordinates=_construct_coord_indices_str(inf_indices, n_top_indices)
                        ),
                        inner_exception=ve
                    ) from ve


def _check_all_finite(data: DataInputType, input_data_name: str) -> None:
    """
    Calls in to sklearn's assert_all_finite() implementation. A ValueError is raised if the check fails.
    An AttributeError is raised if the input data is malformed.

    If a dataset failed sklearn validation, it cannot be trained by most of our pipeline.
    Reference: https://github.com/scikit-learn/scikit-learn/blob/0.19.X/sklearn/utils/validation.py)
    """
    try:
        sk_validation.assert_all_finite(data)
    except AttributeError as ae:
        logger.error("Failed to validate whether the input Dataset {} contains all finite values. "
                     "Cannot pre-process the data.".format(input_data_name))
        raise DataException._with_error(
            AzureMLError.create(InvalidValuesInData, target=input_data_name),
            inner_exception=ae
        ) from ae


def _construct_coord_indices_str(data_indices: np.ndarray, n_top_indices: int = 20) -> str:
    """Contruct a string with top 20 indices."""
    if data_indices.ndim == 1 or data_indices.shape[1] == 1:
        indices = sorted(data_indices)
    else:
        indices = sorted(data_indices, key=lambda x: (x[1], x[0]))
    if len(indices) <= n_top_indices:
        print_indices = data_indices  # type: Union[np.ndarray, List[List[int]]]
        return ", ".join([str(idx) for idx in print_indices])
    else:
        if data_indices.ndim == 1:
            print_indices = data_indices[:n_top_indices]
        else:
            col_idx_dict = {}  # type: Dict[int, List[List[int]]]
            for idx in indices:
                if idx[1] not in col_idx_dict:
                    col_idx_dict[idx[1]] = [idx]
                else:
                    col_idx_dict[idx[1]].append(idx)
            top_indices = sorted(col_idx_dict.keys(), key=lambda x: len(col_idx_dict[x]))
            if len(top_indices) > n_top_indices:
                print_indices = [col_idx_dict[idx][0] for idx in top_indices[:n_top_indices]]
            else:
                print_indices = [col_idx_dict[idx][0] for idx in top_indices]
        return ", ".join([str(idx) for idx in print_indices]) + "..."


def _get_data_indices_by_mask_function(data: CoreDataInputType,
                                       mask_function: 'Callable[..., Optional[Any]]') -> np.ndarray:
    """Obtain the indices list where the data entry in data has the mask function evaluated as True."""
    if isinstance(data, np.ndarray) or isinstance(data, pd.DataFrame):
        return np.argwhere(mask_function(data))
    elif scipy.sparse.issparse(data):
        coo_data = scipy.sparse.coo_matrix(data)
        return np.array(
            [(coo_data.row[i], coo_data.col[i]) for i in np.argwhere(mask_function(coo_data.data)).ravel()])
    else:
        return np.array([])


def check_sample_weight(x: DataInputType,
                        sample_weight: np.ndarray,
                        automl_settings: AutoMLBaseSettings,
                        is_validation_data: bool = False) -> None:
    """
    Validate sample_weight.

    :param x:
    :param sample_weight:
    :param sample_weight_name:
    :param automl_settings:
    :raise DataException if sample_weight has problems
    :return:
    """
    sample_weight_name = "sample_weight_valid" if is_validation_data else "sample_weight"
    Validation.validate_type(sample_weight, sample_weight_name, expected_types=np.ndarray,
                             reference_code=ReferenceCodes._VALIDATE_SAMPLE_WEIGHTS_NDARRAY)

    if x.shape[0] != len(sample_weight):
        raise DataException._with_error(AzureMLError.create(
            SampleCountMismatch, target="X/sample_weight",
            first_dataset_name="X", first_dataset_shape=x.shape[0],
            second_dataset_name=sample_weight_name, second_dataset_shape=len(sample_weight),
            reference_code=ReferenceCodes._VALIDATE_SAMPLE_WEIGHTS_NUM_RECORD)
        )

    if len(sample_weight.shape) > 1:
        raise DataException._with_error(AzureMLError.create(
            DataShapeMismatch, target=sample_weight_name, data_object_name="y", expected_dim=1,
            actual_dim=len(sample_weight.shape), reference_code=ReferenceCodes._VALIDATE_SAMPLE_WEIGHTS_SHAPE)
        )

    if automl_settings.primary_metric in \
            constants.Metric.SAMPLE_WEIGHTS_UNSUPPORTED_SET:
        raise ValidationException._with_error(AzureMLError.create(
            SampleWeightsUnsupported, target=sample_weight_name,
            primary_metrics=", ".join(constants.Metric.SAMPLE_WEIGHTS_UNSUPPORTED_SET),
            reference_code=ReferenceCodes._VALIDATE_SAMPLE_WEIGHTS_METRIC)
        )

    sample_weight_dtype = runtime_utilities._get_column_data_type_as_str(sample_weight)
    if not utilities._check_if_column_data_type_is_numerical(sample_weight_dtype):
        raise ValidationException._with_error(AzureMLError.create(
            InvalidArgumentType, target=sample_weight_name, argument=sample_weight_name,
            actual_type=sample_weight_dtype, expected_types=", ".join(list(NumericalDtype.FULL_SET)),
            reference_code=ReferenceCodes._VALIDATE_SAMPLE_WEIGHTS_IS_NUMERIC)
        )

    sample_weight_min = np.amin(sample_weight)
    if sample_weight_min < 0:
        raise ValidationException._with_error(
            AzureMLError.create(
                ArgumentOutOfRange, target="sample_weight_min",
                argument_name="sample_weight_min ({})".format(str(sample_weight_min)), min=0, max="inf",
                reference_code=ReferenceCodes._VALIDATE_SAMPLE_WEIGHTS_IN_RANGE
            )
        )


def _validate_featurization_config(featurization: Union[str, Dict[str, Any]],
                                   x_raw_column_name: Optional[np.ndarray]) -> None:
    """
    Check if columns with custom purpose or featurization are present in the column list.

    :param featurization: The featurization config object.
    :param x_raw_column_name: the data frame column names.
    :raises: ConfigException
    """
    if x_raw_column_name is None:
        return
    if isinstance(featurization, FeaturizationConfig):
        if featurization.column_purposes is not None:
            for col in featurization.column_purposes.keys():
                if col not in x_raw_column_name:
                    raise ConfigException._with_error(
                        AzureMLError.create(
                            FeaturizationConfigColumnMissing, target='X', columns=col,
                            sub_config_name="column_purposes", all_columns=x_raw_column_name,
                            reference_code=ReferenceCodes._VALIDATE_FEATURIZATION_PURPOSE_TIMESERIES
                        )
                    )
        if featurization.transformer_params is not None:
            for _, col_param_list in featurization.transformer_params.items():
                for col_param in col_param_list:
                    if col_param[0] not in x_raw_column_name:
                        raise ConfigException._with_error(
                            AzureMLError.create(
                                FeaturizationConfigColumnMissing, target='X', columns=col_param[0],
                                sub_config_name="transformer_params", all_columns=x_raw_column_name,
                                reference_code=ReferenceCodes._VALIDATE_FEATURIZATION_TRANSFORM_TIMESERIES
                            )
                        )


def _check_dimensions(X: DataInputType,
                      y: DataInputType,
                      X_valid: DataInputType,
                      y_valid: DataInputType,
                      sample_weight: DataInputType,
                      sample_weight_valid: DataInputType) -> None:
    """
    Check dimensions of data inputs.

    :param X: Training Data
    :param y: Labels
    :param X_valid: Validation Data
    :param y_valid: Validation Labels
    :param sample_weight: Training sample weights
    :param sample_weight_valid: Validation sample weights
    :return: None
    """
    feature_dimensions = 2
    label_dimensions = 1

    # if the data is not in these 4 type, we will bypass the test
    x_dim = None
    y_dim = None
    x_valid_dim = None
    y_valid_dim = None
    sample_weight_dim = None
    sample_weight_valid_dim = None
    x_shape = None
    y_shape = None
    x_valid_shape = None
    y_valid_shape = None
    sample_weight_shape = None
    sample_weight_valid_shape = None

    if X is not None:
        if isinstance(X, pd.DataFrame) or isinstance(X, np.ndarray) or scipy.sparse.issparse(X):
            x_shape = X.shape
            x_dim = X.ndim
        elif isinstance(X, dprep.Dataflow):
            # calculating shape on Dataflows can be very expensive, as it calculates the profile of the underlying data
            # we'd probably need a different design to do this check on dataflows
            x_shape = X.shape
            x_dim = x_shape[1]
        else:
            raise ValidationException._with_error(
                AzureMLError.create(
                    InvalidInputDatatype, target="X", input_type=type(X),
                    supported_types=", ".join(SupportedInputDatatypes.ALL)
                )
            )

    if X_valid is not None:
        if isinstance(X_valid, pd.DataFrame) or isinstance(X_valid, np.ndarray) or scipy.sparse.issparse(X_valid):
            x_valid_shape = X_valid.shape
            x_valid_dim = X_valid.ndim
        elif isinstance(X_valid, dprep.Dataflow):
            x_valid_shape = X_valid.shape
            x_valid_dim = x_valid_shape[1]
        else:
            raise ValidationException._with_error(
                AzureMLError.create(
                    InvalidInputDatatype, target="X_valid", input_type=type(X_valid),
                    supported_types=", ".join(SupportedInputDatatypes.ALL)
                )
            )

    if y is not None:
        if isinstance(y, pd.DataFrame) or scipy.sparse.issparse(y) or isinstance(y, dprep.Dataflow):
            y_shape = y.shape
            y_dim = y.shape[1]
        elif isinstance(y, np.ndarray):
            y_shape = y.shape
            y_dim = y.ndim
        else:
            raise ValidationException._with_error(
                AzureMLError.create(
                    InvalidInputDatatype, target="y", input_type=type(y),
                    supported_types=", ".join(SupportedInputDatatypes.ALL)
                )
            )

    if y_valid is not None:
        if isinstance(y_valid, pd.DataFrame) or scipy.sparse.issparse(y_valid) or isinstance(y_valid, dprep.Dataflow):
            y_valid_shape = y_valid.shape
            y_valid_dim = y_valid.shape[1]
        elif isinstance(y_valid, np.ndarray):
            y_valid_shape = y_valid.shape
            y_valid_dim = y_valid.ndim
        else:
            raise ValidationException._with_error(
                AzureMLError.create(
                    InvalidInputDatatype, target="y_valid", input_type=type(y_valid),
                    supported_types=", ".join(SupportedInputDatatypes.ALL)
                )
            )

    if sample_weight is not None:
        if isinstance(sample_weight, pd.DataFrame) or \
                scipy.sparse.issparse(sample_weight) or \
                isinstance(sample_weight, dprep.Dataflow):
            sample_weight_shape = sample_weight.shape
            sample_weight_dim = sample_weight.shape[1]
        elif isinstance(sample_weight, np.ndarray):
            sample_weight_shape = sample_weight.shape
            sample_weight_dim = sample_weight.ndim
        else:
            raise ValidationException._with_error(
                AzureMLError.create(
                    InvalidInputDatatype, target="sample_weight", input_type=type(sample_weight),
                    supported_types=", ".join(SupportedInputDatatypes.ALL)
                )
            )

    if sample_weight_valid is not None:
        if isinstance(sample_weight_valid, pd.DataFrame) or \
                scipy.sparse.issparse(sample_weight_valid) or \
                isinstance(sample_weight_valid, dprep.Dataflow):
            sample_weight_valid_shape = sample_weight_valid.shape
            sample_weight_valid_dim = sample_weight_valid.shape[1]
        elif isinstance(sample_weight_valid, np.ndarray):
            sample_weight_valid_shape = sample_weight_valid.shape
            sample_weight_valid_dim = sample_weight_valid.ndim
        else:
            raise ValidationException._with_error(
                AzureMLError.create(
                    InvalidInputDatatype, target="sample_weight_valid", input_type=type(sample_weight_valid),
                    supported_types=", ".join(SupportedInputDatatypes.ALL)
                )
            )

    if x_dim is not None and x_dim > feature_dimensions:
        raise DataException._with_error(AzureMLError.create(
            DataShapeMismatch, target="X", data_object_name="X", expected_dim=feature_dimensions,
            actual_dim=x_dim)
        )
    if y_dim is not None and y_dim != label_dimensions:
        raise DataException._with_error(AzureMLError.create(
            DataShapeMismatch, target="y", data_object_name="y", expected_dim=label_dimensions,
            actual_dim=y_dim)
        )
    if x_valid_dim is not None and x_valid_dim > feature_dimensions:
        raise DataException._with_error(AzureMLError.create(
            DataShapeMismatch, target="X_valid", data_object_name="X_valid", expected_dim=feature_dimensions,
            actual_dim=x_valid_dim)
        )
    if y_valid_dim is not None and y_valid_dim != label_dimensions:
        raise DataException._with_error(AzureMLError.create(
            DataShapeMismatch, target="y_valid", data_object_name="y_valid", expected_dim=label_dimensions,
            actual_dim=y_valid_dim)
        )
    if sample_weight_dim is not None and sample_weight_dim != label_dimensions:
        raise DataException._with_error(AzureMLError.create(
            DataShapeMismatch, target="sample_weight", data_object_name="sample_weight", expected_dim=label_dimensions,
            actual_dim=sample_weight_dim)
        )
    if sample_weight_valid_dim is not None and sample_weight_valid_dim != label_dimensions:
        raise DataException._with_error(AzureMLError.create(
            DataShapeMismatch, target="sample_weight_valid", data_object_name="sample_weight_valid",
            expected_dim=label_dimensions, actual_dim=sample_weight_valid_dim)
        )

    if x_shape is not None and y_shape is not None and x_shape[0] != y_shape[0]:
        raise DataException._with_error(AzureMLError.create(
            SampleCountMismatch, target="X/y",
            first_dataset_name="X", first_dataset_shape=x_shape[0],
            second_dataset_name="y", second_dataset_shape=y_shape[0])
        )
    if x_valid_shape is not None and y_valid_shape is not None and \
            x_valid_shape[0] != y_valid_shape[0]:
        raise DataException._with_error(AzureMLError.create(
            SampleCountMismatch, target="X_valid/y_valid",
            first_dataset_name="X_valid", first_dataset_shape=x_valid_shape[0],
            second_dataset_name="y_valid", second_dataset_shape=y_valid_shape[0])
        )
    if sample_weight_shape is not None and y_shape is not None and \
            sample_weight_shape[0] != y_shape[0]:
        raise DataException._with_error(AzureMLError.create(
            SampleCountMismatch, target="sample_weight/y",
            first_dataset_name="sample_weight", first_dataset_shape=sample_weight_shape[0],
            second_dataset_name="y", second_dataset_shape=y_shape[0])
        )
    if sample_weight_valid_shape is not None and y_valid_shape is not None and\
            sample_weight_valid_shape[0] != y_valid_shape[0]:
        raise DataException._with_error(AzureMLError.create(
            SampleCountMismatch, target="sample_weight_valid/y_valid",
            first_dataset_name="sample_weight_valid", first_dataset_shape=sample_weight_valid_shape[0],
            second_dataset_name="y_valid", second_dataset_shape=y_valid_shape[0])
        )
    if x_shape is not None and y_shape is not None and x_shape[0] == 0:
        raise DataException._with_error(AzureMLError.create(InputDatasetEmpty, target="X"))
    if x_valid_shape is not None and y_valid_shape is not None and x_valid_shape[0] == 0:
        raise DataException._with_error(AzureMLError.create(InputDatasetEmpty, target="X_valid"))


def _is_sparse_matrix_int_type(sparse_matrix: DataInputType) -> bool:
    """
    Check if a sparse matrix is in integer format.

    :param sparse_matrix:
    :return:
    """
    if sparse_matrix is not None and scipy.sparse.issparse(sparse_matrix):
        numpy_int_types = [np.int32, np.int64, np.int16, np.int8,
                           np.uint32, np.uint64, np.uint16, np.uint8]

        if sparse_matrix.dtype in numpy_int_types:
            return True

    return False


def _upgrade_sparse_matrix_type(sparse_matrix: DataInputType) -> DataInputType:
    """
    Convert sparse matrix in integer format into floating point format.

    This function will create a copy of the sparse matrix in when the conversion happens.
    :param sparse_matrix:
    :return:
    """
    if sparse_matrix is not None and scipy.sparse.issparse(sparse_matrix):
        if sparse_matrix.dtype == np.int32 or sparse_matrix.dtype == np.int16 or \
                sparse_matrix.dtype == np.int8 or sparse_matrix.dtype == np.uint32 or \
                sparse_matrix.dtype == np.uint16 or sparse_matrix.dtype == np.uint8 or \
                sparse_matrix.dtype == np.bool:
            return sparse_matrix.astype(np.float32)
        elif sparse_matrix.dtype == np.int64 or sparse_matrix.dtype == np.uint64:
            return sparse_matrix.astype(np.float64)
        else:
            return sparse_matrix

    return sparse_matrix


def init_client_dataset_from_fit_iteration_params(fit_iteration_parameters_dict: Dict[str, Any],
                                                  automl_settings: AutoMLBaseSettings,
                                                  cache_store: Optional[CacheStore] = None,
                                                  remote: bool = False,
                                                  init_all_stats: bool = False,
                                                  keep_in_memory: bool = False) -> ClientDatasets:
    """
    Get a ClientDatasets object from fit_iteration_parameters

    TODO: This method needs to be deprecated. ClientDatasets should be consolidated to only use transformed data ctx

    :param fit_iteration_parameters_dict: Dictionary that contains input data
    :param automl_settings:  AutoML settings config
    :param cache_store: Underlying cache store to use, will default to local FileStore
    :param remote: remote or local run flag
    :param init_all_stats: Initialize all the stats
    :param keep_in_memory: Whether to flush the data to the cache store or keep it in-memory
    :return: ClientDatasets
    """
    cv_splits = _CVSplits(X=fit_iteration_parameters_dict.get('X'),
                          y=fit_iteration_parameters_dict.get('y'),
                          frac_valid=automl_settings.validation_size,
                          cv_splits_indices=fit_iteration_parameters_dict.get('cv_splits_indices'),
                          is_time_series=automl_settings.is_timeseries,
                          timeseries_param_dict=utilities._get_ts_params_dict(automl_settings))

    dataset = _get_client_dataset(fit_iteration_parameters_dict.get('X'),
                                  fit_iteration_parameters_dict.get('y'),
                                  cache_store=cache_store,
                                  sample_weight=fit_iteration_parameters_dict.get('sample_weight'),
                                  X_valid=fit_iteration_parameters_dict.get('X_valid'),
                                  y_valid=fit_iteration_parameters_dict.get('y_valid'),
                                  sample_weight_valid=fit_iteration_parameters_dict.get('sample_weight_valid'),
                                  cv_splits=cv_splits,
                                  num_classes=automl_settings.num_classes,
                                  task_type=automl_settings.task_type,
                                  y_min=automl_settings.y_min,
                                  y_max=automl_settings.y_max,
                                  init_all_stats=init_all_stats,
                                  remote=remote)

    dataset.x_raw_column_names = fit_iteration_parameters_dict.get('x_raw_column_names')

    if not keep_in_memory:
        dataset.cache_dataset(keep_in_memory)

    return dataset


def init_dataset(
    transformed_data_context: Union[TransformedDataContext, StreamingTransformedDataContext],
    cache_store: CacheStore,
    automl_settings: AutoMLBaseSettings,
    remote: bool = False,
    init_all_stats: bool = False,
    keep_in_memory: bool = False
) -> DatasetBase:
    """
    Initialize the dataset.

    :param transformed_data_context: transformed_data_context contains X,y & other data's.
    :param cache_store: cache store
    :param automl_settings: automl settings
    :param remote: remote or local run flag
    :param init_all_stats: init all stats
    :param keep_in_memory: Whether to flush the data to the cache store or keep it in-memory
    :return: DatasetBase
    """
    if isinstance(transformed_data_context, StreamingTransformedDataContext):
        return init_streaming_dataset(
            transformed_data_context=transformed_data_context,
            automl_settings=automl_settings
        )

    elif isinstance(transformed_data_context, TransformedDataContext):
        return init_client_dataset(
            transformed_data_context=transformed_data_context,
            cache_store=cache_store,
            automl_settings=automl_settings,
            remote=remote,
            init_all_stats=init_all_stats,
            keep_in_memory=keep_in_memory)


def init_client_dataset(transformed_data_context: TransformedDataContext,
                        cache_store: CacheStore,
                        automl_settings: AutoMLBaseSettings,
                        remote: bool = False,
                        init_all_stats: bool = False,
                        keep_in_memory: bool = False) -> ClientDatasets:
    """
    Get the client dataset.

    :param transformed_data_context: transformed_data_context contains X,y & other data's.
    :param cache_store: cache store
    :param automl_settings: automl settings
    :param remote: remote or local run flag
    :param init_all_stats: init all stats
    :param keep_in_memory: Whether to flush the data to the cache store or keep it in-memory
    :return: ClientDatasets
    """

    dataset = _get_client_dataset(transformed_data_context.X,
                                  transformed_data_context.y,
                                  cache_store=cache_store,
                                  sample_weight=transformed_data_context.sample_weight,
                                  X_valid=transformed_data_context.X_valid,
                                  y_valid=transformed_data_context.y_valid,
                                  sample_weight_valid=transformed_data_context.sample_weight_valid,
                                  cv_splits=transformed_data_context.cv_splits,
                                  num_classes=automl_settings.num_classes,
                                  task_type=automl_settings.task_type,
                                  y_min=automl_settings.y_min,
                                  y_max=automl_settings.y_max,
                                  init_all_stats=init_all_stats,
                                  remote=remote,
                                  transformers=transformed_data_context.transformers)
    dataset.timeseries = transformed_data_context.timeseries
    dataset.timeseries_param_dict = transformed_data_context.timeseries_param_dict
    dataset.x_raw_column_names = transformed_data_context.x_raw_column_names
    dataset.raw_data_type = transformed_data_context._get_raw_data_type()
    dataset.raw_data_snapshot_str = transformed_data_context._get_raw_data_snapshot_str()

    if automl_settings.n_cross_validations is None and transformed_data_context.X_valid is None:
        # set the value for num_auto_cv_splits if no other mode of Validation is specified
        n_cv = transformed_data_context._get_num_cv_splits()
        dataset.num_auto_cv_splits = None if n_cv == 0 else n_cv

    if not keep_in_memory:
        dataset.cache_dataset(keep_in_memory)

    return dataset


def _get_client_dataset(X: DataInputType,
                        y: DataSingleColumnInputType,
                        cache_store: Optional[CacheStore] = None,
                        sample_weight: Optional[DataInputType] = None,
                        X_valid: Optional[DataInputType] = None,
                        y_valid: Optional[DataSingleColumnInputType] = None,
                        sample_weight_valid: Optional[DataInputType] = None,
                        cv_splits: Optional[_CVSplits] = None,
                        num_classes: Optional[int] = None,
                        task_type: str = constants.Tasks.CLASSIFICATION,
                        y_min: Optional[float] = None,
                        y_max: Optional[float] = None,
                        init_all_stats: bool = False,
                        remote: bool = True,
                        transformers: Optional[Dict[str, TransformerMixin]] = None) -> ClientDatasets:
    default_dataset_name = 'NoName'

    if cv_splits:
        frac_valid = cv_splits.get_fraction_validation_size()
        cv_splits_indices = cv_splits.get_custom_split_indices()
        num_cv_folds = cv_splits.get_num_k_folds()
    else:
        frac_valid = None
        cv_splits_indices = None
        num_cv_folds = None

    subsample_cache_strategy = SubsampleCacheStrategy.Classic if remote \
        else SubsampleCacheStrategy.Preshuffle

    dataset = ClientDatasets(subsample_cache_strategy=subsample_cache_strategy, cache_store=cache_store)
    logger.info('Created ClientDataset.')

    if X_valid is not None:
        training_type = _get_training_type(
            constants.TrainingType.TrainAndValidation)

        if not (num_cv_folds == 0 or num_cv_folds is None):
            raise ConfigException._with_error(
                AzureMLError.create(
                    ConflictingValueForArguments, target="n_cross_validations",
                    arguments=', '.join(['validation_data/X_valid', 'n_cross_validations'])
                )
            )

        if not (frac_valid == 0.0 or frac_valid is None):
            raise ConfigException._with_error(
                AzureMLError.create(
                    ConflictingValueForArguments, target="validation_size",
                    arguments=', '.join(['validation_data/X_valid', 'validation_size'])
                )
            )

        if y_valid is None:
            raise ConfigException._with_error(
                AzureMLError.create(
                    ArgumentBlankOrEmpty, target="y_valid", argument_name="y_valid"
                )
            )

        logger.info('Parsing simple train validate dataset.')
        dataset.parse_simple_train_validate(name=default_dataset_name,
                                            X=X,
                                            y=y,
                                            sample_weight=sample_weight,
                                            X_valid=X_valid,
                                            y_valid=y_valid,
                                            sample_weight_valid=sample_weight_valid,
                                            task=task_type,
                                            y_min=y_min,
                                            y_max=y_max,
                                            init_all_stats=init_all_stats,
                                            transformers=transformers)

    else:
        if (num_cv_folds == 0 or num_cv_folds is None) and cv_splits_indices is None:
            training_type = _get_training_type(
                constants.TrainingType.TrainAndValidation)
        else:
            if cv_splits_indices is not None:
                num_cv_folds = len(cv_splits_indices)
            training_type = _get_training_type(
                constants.TrainingType.MeanCrossValidation, num_cv_folds)

        logger.info('Parsing cross validation dataset.')
        dataset.parse_data(name=default_dataset_name,
                           X=X,
                           y=y,
                           sample_weight=sample_weight,
                           cv_splits=cv_splits,
                           num_classes=num_classes,
                           task=task_type,
                           y_min=y_min,
                           y_max=y_max,
                           init_all_stats=init_all_stats,
                           transformers=transformers)

    dataset.training_type = training_type

    return dataset


def init_streaming_dataset(
    transformed_data_context: StreamingTransformedDataContext,
    automl_settings: AutoMLBaseSettings
) -> StreamingDataset:
    """
    Initialize a streaming dataset (a dataset where where all data may not fit into memory at once).

    :param transformed_data_context: The transformed data context.
    :param automl_settings: AutoML settings
    :return: A StreamingDataset
    """
    if automl_settings.label_column_name is None:
        raise ConfigException._with_error(
            AzureMLError.create(
                ArgumentBlankOrEmpty, target="label_column_name", argument_name="label_column_name"
            )
        )

    featurized_column_names = transformed_data_context.get_featurized_column_names()

    dataset_metadata = {DatasetMetadataKeys.feature_column_names: featurized_column_names,
                        DatasetMetadataKeys.label_column_name: automl_settings.label_column_name,
                        DatasetMetadataKeys.weight_column_name: automl_settings.weight_column_name,
                        DatasetMetadataKeys.raw_data_snapshot: transformed_data_context.raw_data_snapshot}

    featurization_transformer = transformed_data_context.get_featurization_transformer()

    return StreamingDataset(task=automl_settings.task_type,
                            training_data=transformed_data_context.training_data,
                            dataset_metadata=dataset_metadata,
                            validation_data=transformed_data_context.validation_data,
                            y_min=automl_settings.y_min,
                            y_max=automl_settings.y_max,
                            featurization_transformer=featurization_transformer)


def _get_training_type(training_type: str, folds: int = 0) -> str:
    """
    Determine what type of training and validation to do based on user inputs.
    """
    # TODO: make this simpler
    valid_training_types = (constants.TrainingType.TrainAndValidation,
                            constants.TrainingType.MeanCrossValidation)
    if training_type not in valid_training_types:
        raise ConfigException._with_error(
            AzureMLError.create(
                InvalidArgumentWithSupportedValues, target="training_type",
                arguments="training_type ({})".format(training_type), supported_values="%s, %s" % valid_training_types
            )
        )
    is_cv = training_type == constants.TrainingType.MeanCrossValidation
    if not ((is_cv and folds) or (not is_cv and not folds)):
        # This isn't very user friendly / actionable.
        raise ConfigException._with_error(
            AzureMLError.create(
                ConflictingValueForArguments, target="cross_validation_folds",
                arguments=', '.join(['training_type ({})'.format(constants.TrainingType.MeanCrossValidation),
                                     'cross_validation_folds'])
            )
        )
    if folds < 0 or folds == 1:
        raise ConfigException._with_error(
            AzureMLError.create(
                ArgumentOutOfRange, target="cross_validation_folds",
                argument_name="cross_validation_folds", min=2, max="inf"
            )
        )
    return training_type


def _extract_user_data(user_script: Any) -> Dict[str, Optional[Union[np.ndarray, List[str], float, List[int]]]]:
    """
    Extract data from user's module containing get_data().

    This method automatically runs during an automated machine learning experiment.
    Arguments:
        user_script {module} -- Python module containing get_data() function.

    Raises:
        DataException -- Get data script was not defined and X, y inputs were not provided.
        DataException -- Could not execute get_data() from user script.
        DataException -- Could not extract data from user script.

    Returns:
        dict -- Dictionary containing
        X_train, y_train, sample_weight, X_valid, y_valid,
        sample_weight_valid, cv_splits_indices.

    """
    Validation.validate_value(user_script, "user_script")
    try:
        output = user_script.get_data()         # type: Union[Dict[str, Any], Tuple[Any, Any, Any, Any]]
    except Exception as ex:
        msg = ("Could not execute get_data() from user script."
               "Exception: {}")
        raise UserException(msg.format(ex)).with_generic_msg(msg.format(MASKED)) from None
    if isinstance(output, dict):
        return _extract_data_from_dict(output)
    elif isinstance(output, tuple):
        return _extract_data_from_tuple(output)
    else:
        raise DataException("The output of get_data() from user script is not a dict or a tuple.", has_pii=False)


def _extract_data_from_dict(output: Dict[str, Any]) -> \
        Dict[str, Optional[Union[np.ndarray, List[str], float, List[int]]]]:
    """
    Extract user data if it is passed as a dictionary.

    Arguments:
        output {dict} -- dictionary containing user data and metadata.

    Raises:
        DataException -- Invalid data or encountered processing issues.

    Returns:
        dict -- Dictionary containing AutoML relevant data.

    """
    X = utilities.get_value_from_dict(output, ['X'], None)
    y = utilities.get_value_from_dict(output, ['y'], None)
    sample_weight = utilities.get_value_from_dict(output, ['sample_weight'], None)
    X_valid = utilities.get_value_from_dict(output, ['X_valid'], None)
    y_valid = utilities.get_value_from_dict(output, ['y_valid'], None)
    sample_weight_valid = utilities.get_value_from_dict(
        output, ['sample_weight_valid'], None)
    X_test = utilities.get_value_from_dict(output, ['X_test'], None)
    y_test = utilities.get_value_from_dict(output, ['y_test'], None)
    data = utilities.get_value_from_dict(output, ['data_train'], None)
    columns = utilities.get_value_from_dict(output, ['columns'], None)
    label = utilities.get_value_from_dict(output, ['label'], None)
    cv_splits_indices = utilities.get_value_from_dict(
        dictionary=output,
        names=["cv_splits_indices"], default_value=None)
    x_raw_column_names = None

    if data is not None:
        if label is None and X is None and y is None:
            # This is just called in the get_data code path.
            # get_data is to soon expected to be deprecated (and has non-existent usage among customers)
            raise DataException(
                'Pandas data array received without a label. Please add a ''label'' element to the '
                'get_data() output.', has_pii=False)
        if not isinstance(label, list):
            assert(isinstance(label, str) or isinstance(label, int))
            label = [label]
        y_extracted = data[label].values
        X_extracted = data[data.columns.difference(label)]
        if columns is not None:
            X_extracted = X_extracted[X_extracted.columns.intersection(
                columns)]

        if X is None and y is None:
            X = X_extracted
            y = y_extracted
        else:
            if np.array_equiv(X, X_extracted.values):
                raise DataException(
                    "Different values for X and data were provided. "
                    "Please return either X and y or data and label.", has_pii=False)
            if np.array_equiv(y, y_extracted.values):
                raise DataException(
                    "Different values for y and label were provided. "
                    "Please return either X and y or data and label.", has_pii=False)
    if isinstance(X, pd.DataFrame):
        x_raw_column_names = X.columns.values
        X = X.values
    if isinstance(X_valid, pd.DataFrame):
        X_valid = X_valid.values
    if isinstance(X_test, pd.DataFrame):
        X_test = X_test.values
    if isinstance(y, pd.DataFrame):
        y = y.values
    if isinstance(y_valid, pd.DataFrame):
        y_valid = y_valid.values
    if isinstance(y_test, pd.DataFrame):
        y_test = y_test.values

    if X is None:
        raise DataException(
            "Could not retrieve X train data from get_data() call. "
            "Please ensure you are either returning either "
            "{X_train: <numpy array>, y_train: <numpy array>"
            "or {data: <pandas dataframe>, label: <string>", has_pii=False)
    if y is None:
        raise DataException(
            "Could not retrieve y train data from get_data() call. "
            "Please ensure you are either returning either "
            "{X_train: <numpy array>, y_train: <numpy array>"
            "or {data: <pandas dataframe>, label: <string>", has_pii=False)

    if (X_valid is None) is not (y_valid is None):
        raise DataException(
            'Received only one of X_valid or y_valid.'
            'Either both or neither value should be provided.', has_pii=False)

    return {
        "X": X,
        "y": y,
        "x_raw_column_names": x_raw_column_names,
        "sample_weight": sample_weight,
        "X_valid": X_valid,
        "y_valid": y_valid,
        "sample_weight_valid": sample_weight_valid,
        "X_test": X_test,
        "y_test": y_test,
        "cv_splits_indices": cv_splits_indices,
    }


def _extract_data_from_tuple(
        output: Tuple[Union[pd.DataFrame, np.ndarray], Union[pd.DataFrame, np.ndarray],
                      Union[pd.DataFrame, np.ndarray], Union[pd.DataFrame, np.ndarray]]) \
        -> Dict[str, Optional[Union[np.ndarray, List[str], float, List[int]]]]:
    """
    Extract user data if it is passed as a tuple.

    Arguments:
        output {tuple} -- tuple containing user data.

    Raises:
        DataException -- Could not extract X, y from get_data() in user script. get_data only output {0} values.

    Returns:
        tuple -- tuple containing X_train, y_train, X_test, y_test

    """
    X_valid, y_valid = None, None
    if len(output) < 2:
        msg = ("Could not extract X, y from get_data() in user "
               "script. get_data only output {0} values.")
        raise DataException(msg.format(len(output))).with_generic_msg(msg.format(MASKED)) from None
    x_raw_column_names = None
    X = output[0]
    y = output[1]
    if isinstance(X, pd.DataFrame):
        x_raw_column_names = X.columns.values
        X = X.values
    if isinstance(y, pd.DataFrame):
        y = y.values

    if len(output) >= 4:
        X_valid = output[2]
        y_valid = output[3]
        if isinstance(y_valid, pd.DataFrame):
            y_valid = y_valid.values
        if isinstance(X_valid, pd.DataFrame):
            X_valid = X_valid.values

    return {
        "X": X,
        "y": y,
        "sample_weight": None,
        "x_raw_column_names": x_raw_column_names,
        "X_valid": X_valid,
        "y_valid": y_valid,
        "sample_weight_valid": None,
        "X_test": None,
        "y_test": None,
        "cv_splits_indices": None,
    }


def _extract_data_from_combined_dataframe(
        training_data: pd.DataFrame,
        label_column_name: str,
        sample_weight_column_name: Optional[str] = None,
        cv_split_column_names: Optional[List[str]] = None
) -> Tuple[Any, Any, Any, Any]:
    """
    Extract user data from a Pandas dataframe if it contains both training features & labels.

    :param training_data: The Pandas dataframe to extract X, y, sample_valid from.
    :param label_column_name: Name of the label column used to extract y.
    :param sample_weight_column_name: Name of the sample weight column.
    :param cv_split_column_names: List of names of the cv split columns.
    :return: A Dictionary with keys being X, y, sample_weight, cv_splits_indices extracted from training_data.
    """
    col_names_to_drop = [label_column_name]
    sample_weight = None
    cv_splits_indices = None

    if sample_weight_column_name is not None:
        col_names_to_drop.append(sample_weight_column_name)
        if sample_weight_column_name not in training_data.columns:
            raise DataException._with_error(AzureMLError.create(
                MissingColumnsInData, target='sample_weight_column_name', columns=sample_weight_column_name,
                data_object_name='training_data',
                reference_code=ReferenceCodes._EXTRACT_DATA_FROM_COMBINED_DATAFRAME_SAMPLE_WEIGHT_MISSING)
            )
        sample_weight = training_data[sample_weight_column_name].values

    if cv_split_column_names is not None:
        col_names_to_drop.extend(cv_split_column_names)
        for cv_split_column_name in cv_split_column_names:
            if cv_split_column_name not in training_data.columns:
                raise DataException._with_error(AzureMLError.create(
                    MissingColumnsInData, target='cv_split_column_names', columns=cv_split_column_name,
                    data_object_name='training_data',
                    reference_code=ReferenceCodes._EXTRACT_DATA_FROM_COMBINED_DATAFRAME_CV_SPLIT_COLUMNS_MISSING)
                )
        cv_split_columns = training_data[cv_split_column_names]
        cv_splits_indices = _convert_cv_split_columns_to_cv_splits_indices(cv_split_columns)

    X = training_data[training_data.columns.difference(col_names_to_drop)]
    if label_column_name not in training_data.columns:
        raise DataException._with_error(
            AzureMLError.create(MissingColumnsInData, target='X', columns=label_column_name, data_object_name='X')
        )
    y = training_data[label_column_name].values

    return X, y, sample_weight, cv_splits_indices


def _extract_data_from_combined_dataflow(
        training_data: dprep.Dataflow,
        label_column_name: str,
        sample_weight_column_name: Optional[str] = None,
        cv_split_column_names: Optional[List[str]] = None
) -> Tuple[Any, Any, Any, Any]:
    """
    Extract user data from a Dataflow if it contains both training features & labels.

    :param training_data: The Dataflow to extract X, y, sample_valid from.
    :param label_column_name: Name of the label column used to extract y.
    :param sample_weight_column_name: Name of the sample weight column.
    :param cv_split_column_names: List of names of the cv split columns.
    :return: A Dictionary with keys being X, y, sample_weight, cv_splits_indices extracted from training_data.
    """
    col_names_to_drop = []
    if label_column_name is not None:
        col_names_to_drop.append(label_column_name)
    sample_weight = None
    cv_splits_indices = None

    if sample_weight_column_name is not None:
        col_names_to_drop.append(sample_weight_column_name)
        try:
            sample_weight = training_data.keep_columns([sample_weight_column_name], validate_column_exists=True)
        except DataflowValidationError:
            raise DataException._with_error(AzureMLError.create(
                MissingColumnsInData, target="weight_column_name", data_object_name="X",
                columns="{} (weight_column_name)".format(sample_weight_column_name),
                reference_code=ReferenceCodes._EXTRACT_DATA_FROM_COMBINED_DATAFLOW_SAMPLE_WEIGHT_MISSING)
            )

    if cv_split_column_names is not None:
        cv_split_column_names = list(filter(None, cv_split_column_names))
        col_names_to_drop.extend(cv_split_column_names)
        try:
            cv_split_columns = training_data.keep_columns(cv_split_column_names, validate_column_exists=True)
            cv_splits_indices = _convert_cv_split_columns_to_cv_splits_indices(cv_split_columns)
        except DataflowValidationError:
            raise DataException._with_error(AzureMLError.create(
                MissingColumnsInData, target="cv_split_column_names", data_object_name="X",
                columns="{} (cv_split_column_names)".format(cv_split_column_names),
                reference_code=ReferenceCodes._EXTRACT_DATA_FROM_COMBINED_DATAFLOW_CV_SPLIT_COLUMNS_MISSING)
            )

    X = training_data.drop_columns(col_names_to_drop)
    try:
        y = training_data.keep_columns([label_column_name], validate_column_exists=True) \
            if label_column_name is not None else None
    except DataflowValidationError:
        raise DataException._with_error(AzureMLError.create(
            MissingColumnsInData, target="label_column_name", data_object_name="X",
            columns="{} (label_column_name)".format(label_column_name),
            reference_code=ReferenceCodes._EXTRACT_DATA_FROM_COMBINED_DATAFLOW_LABEL_COLUMN_MISSING)
        )
    return X, y, sample_weight, cv_splits_indices


def _convert_cv_split_columns_to_cv_splits_indices(
        cv_split_columns: Union[dprep.Dataflow, pd.DataFrame]
) -> List[List[Any]]:
    cv_splits_indices = []
    if isinstance(cv_split_columns, pd.DataFrame):
        cv_split_columns_numpy = cv_split_columns.values
    else:
        cv_split_columns_numpy = dataprep_utilities.materialize_dataflow(cv_split_columns, as_numpy=True)
    if cv_split_columns_numpy.ndim == 1:
        training_indices, validation_indices = _get_column_to_cv_splits_indices(cv_split_columns_numpy)
        cv_splits_indices.append([training_indices, validation_indices])
    else:
        for i in range(cv_split_columns_numpy.shape[1]):
            column = cv_split_columns_numpy[:, i]
            training_indices, validation_indices = _get_column_to_cv_splits_indices(column)
            cv_splits_indices.append([training_indices, validation_indices])
    return cv_splits_indices


def _get_column_to_cv_splits_indices(column: np.ndarray) -> Tuple[Any, Any]:
    try:
        training_indices, = np.where(column.astype(int) == 1)
        validation_indices, = np.where(column.astype(int) == 0)

        # verify number of indices extracted equal number of total rows
        if len(training_indices) + len(validation_indices) != len(column):
            raise ValueError("cv split column had values outside of the set {0, 1}")
    except ValueError as ve:
        raise DataException._with_error(
            AzureMLError.create(InvalidValuesInCVSplitColumn, target="cv_split_column_names",
                                reference_code=ReferenceCodes._CV_SPLIT_COLUMN_CONVERSION_INCORRECT_INT_VALUE),
            inner_exception=ve
        ) from ve

    return training_indices, validation_indices


def _check_if_automl_model_is_explainable(automl_algo_tag: str, is_ensemble: bool = False) -> bool:
    unexplainable_models = [constants.SupportedModels.Forecasting.Average,
                            constants.SupportedModels.Forecasting.Naive,
                            constants.SupportedModels.Forecasting.SeasonalAverage,
                            constants.SupportedModels.Forecasting.SeasonalNaive,
                            constants.SupportedModels.Forecasting.TCNForecaster]
    explainable = True
    if is_ensemble:
        # Parse the tag assuming it is a string representation of a Python list
        ensemble_algo_list = ast.literal_eval(automl_algo_tag)
        if len(set(unexplainable_models).intersection(ensemble_algo_list)) > 0:
            explainable = False
    else:
        if automl_algo_tag in unexplainable_models:
            explainable = False

    return explainable


def _get_model_exp_property(automl_run: Any) -> bool:
    try:
        from azureml.train.automl.runtime._remote_script import model_exp_wrapper
        automl_algo_name = automl_run.properties.get('run_algorithm')
        if automl_algo_name != 'StackEnsemble' and automl_algo_name != 'VotingEnsemble':
            if_model_is_explainable = \
                _check_if_automl_model_is_explainable(automl_algo_name, is_ensemble=False)
            if logger is not None and not if_model_is_explainable:
                logger.warning(automl_algo_name + ' is not explainable for AutoML run ' + str(automl_run.id))

            return if_model_is_explainable
        else:
            ensemble_algo_names_list_str = automl_run.tags.get('ensembled_algorithms')
            if ensemble_algo_names_list_str is not None:
                if_ensemble_model_is_explainable = \
                    _check_if_automl_model_is_explainable(ensemble_algo_names_list_str, is_ensemble=True)
                if logger is not None and not if_ensemble_model_is_explainable:
                    msg = ('{} with constituent models {} is not explainable for AutoML run {}'
                           .format(automl_algo_name, ensemble_algo_names_list_str, automl_run.id))
                    logger.warning(msg)

                return if_ensemble_model_is_explainable
            else:
                return True
    except Exception:
        return False


def _check_data_can_be_preprocessed(X: DataInputType,
                                    X_valid: DataInputType,
                                    x_raw_column_names: Optional[np.ndarray] = None) -> None:
    if scipy.sparse.issparse(X):
        return
    n_x_col = 1 if len(X.shape) == 1 else X.shape[1]
    if x_raw_column_names is None and isinstance(X, np.ndarray):
        x_raw_column_names = np.arange(n_x_col)
    elif x_raw_column_names is None:
        # if pandas df, try to use columns_names from dataframe
        x_raw_column_names = X.columns

    for col_num, col_name in zip(range(n_x_col), x_raw_column_names):
        _check_column_can_be_preprocessed(_get_column_by_column_number(X, col_num), col_name, False)
        if X_valid is not None:
            _check_column_can_be_preprocessed(_get_column_by_column_number(X_valid, col_num), col_name, True)


def _check_column_can_be_preprocessed(series: pd.Series, col_name: str, is_valid_data: bool) -> None:
    try:
        # TODO: Check if this is true for all cases
        # preprocess need pandas.unique can be run properly.
        series.unique()
    except TypeError as te:
        input_type = "X_valid" if is_valid_data else "X"
        raise DataException._with_error(AzureMLError.create(
            UnhashableValueInColumn, target=input_type, column_name=col_name, data_object_name=input_type),
            inner_exception=te
        ) from te


def _get_column_by_column_number(X: DataInputType, col_num: int) -> pd.Series:
    if isinstance(X, np.ndarray) and len(X.shape) == 1:
        return pd.Series(X)
    elif isinstance(X, np.ndarray):
        return pd.Series(X[:, col_num])
    else:
        return pd.Series(X.iloc[:, col_num])


def _validate_exp_timeout_for_data(X: DataInputType, automl_settings: AutoMLBaseSettings) -> None:
    if X is not None and automl_settings is not None:
        if automl_settings.experiment_timeout_minutes is not None:
            if not scipy.sparse.issparse(X):
                n_rows = X.shape[0]
                n_cols = 1 if len(X.shape) < 2 else X.shape[1]
                # 1M is the timeout needs to be 60 min
                if n_rows * n_cols > constants.AutoMLValidation.TIMEOUT_DATA_BOUND \
                        and automl_settings.experiment_timeout_minutes < 60:
                    raise ValidationException._with_error(
                        AzureMLError.create(
                            ExperimentTimeoutForDataSize, target="experiment_timeout_minutes",
                            minimum=60, maximum="{:,}".format(constants.AutoMLValidation.TIMEOUT_DATA_BOUND),
                            rows=n_rows, columns=n_cols, total=n_rows * n_cols,
                            reference_code=ReferenceCodes._VALIDATE_EXP_TIMEOUT_WITH_DATA
                        )
                    )


def _validate_all_column_not_ignored(
        X: DataInputType,
        x_raw_column_names: Optional[np.ndarray],
        automl_settings: AutoMLBaseSettings
) -> None:
    """
    Validate whether all columns will be dropped by AutoML or not during featurization.

    :param X: The X input data.
    :param x_raw_column_names: A list of input x column names.
    :param automl_settings: The automl settings.
    """
    if isinstance(X, np.ndarray) or isinstance(X, pd.DataFrame):
        df = _time_series_training_utilities._get_df_or_raise(X, x_raw_column_names, True)
    else:
        # no check for the sparse scenario
        return

    featurization = automl_settings.featurization
    is_customized_featurization_enabled = isinstance(automl_settings.featurization, dict)

    # There is no data transformation when featurization is off.
    if featurization == FeaturizationConfigMode.Off:
        return

    # Featurization auto mode, all the related column set is empty.
    customized_columns = set()  # type: Set[str]
    transformer_params_column_set = set()  # type: Set[str]
    if is_customized_featurization_enabled:
        drop_columns_set = set(featurization.get("_drop_columns") or [])  # type: ignore
        column_purpose_keep_column_dict = {}  # type: Dict[str, str]
        column_purpose_drop_column_dict = {}  # type: Dict[str, str]
        if featurization.get("_column_purposes") is not None:  # type: ignore
            for column, purpose in featurization.get("_column_purposes").items():  # type: ignore
                if purpose in _FeatureType.DROP_SET:
                    column_purpose_drop_column_dict[column] = purpose
                else:
                    column_purpose_keep_column_dict[column] = purpose

        transformer_params = featurization.get("_transformer_params") or {}  # type: ignore
        for transfom_param in transformer_params.values():
            for cols, _ in transfom_param:
                transformer_params_column_set = transformer_params_column_set.union(cols)
        featurization_keep_column_set = set(column_purpose_keep_column_dict.keys())
        featurization_drop_column_set = drop_columns_set.union(column_purpose_drop_column_dict.keys())
        customized_columns = featurization_drop_column_set.union(featurization_keep_column_set)

    stats_and_column_purposes = ColumnPurposeDetector.get_raw_stats_and_column_purposes(df)

    not_auto_dropped_columns = set()  # type: Set[str]
    column_drop_reason_list = []  # type: List[str]
    dropped_transformer_params_column = []  # type: List[Tuple[str, str]]
    for _, feature_type_detected, column in stats_and_column_purposes:
        if feature_type_detected in _FeatureType.DROP_SET and column not in customized_columns:
            column_drop_reason_list.append(
                "Column {} identified as {}.".format(column, feature_type_detected)
            )
            if column in transformer_params_column_set:
                dropped_transformer_params_column.append((column, feature_type_detected))
        else:
            not_auto_dropped_columns.add(column)
    # This should be logged via ConsoleInterface. Commenting this out until ConsoleInterface is available here.
    # if len(dropped_transformer_params_column) > 0:
    #     print("The following transformer_params in featurization config will be ignored as the these columns are"
    #           "automatically dropped by AutoML. If you still want to use these columns, please set correct value"
    #           "in column_purposes of featurizaiton config:")
    #     for col, purpose in dropped_transformer_params_column:
    #         print("Column {}, AutoML detected type {}.".format(col, purpose))

    if len(not_auto_dropped_columns) == 0:
        raise DataException._with_error(AzureMLError.create(
            UnrecognizedFeatures, target="X", column_drop_reasons="\n".join(column_drop_reason_list),
            reference_code=ReferenceCodes._VALIDATE_ALL_COLUMN_IGNORED)
        )

    # Only check this part if featurization config is enabled.
    if is_customized_featurization_enabled:
        final_keep_column = set()  # type: Set[str]
        for column in sorted(not_auto_dropped_columns):
            if column not in featurization_drop_column_set:
                final_keep_column.add(column)
            if column in drop_columns_set:
                column_drop_reason_list.append(
                    "Column {}, included in featurization config's drop columns.".format(column))
            if column in column_purpose_drop_column_dict:
                column_drop_reason_list.append(
                    "Column {}, marked as {} in featurization config.".format(
                        column, column_purpose_drop_column_dict[column]))

        if len(final_keep_column) == 0:
            raise DataException._with_error(AzureMLError.create(
                UnrecognizedFeatures, target="featurization_config",
                column_drop_reasons="\n".join(column_drop_reason_list),
                reference_code=ReferenceCodes._VALIDATE_ALL_COLUMN_IGNORED_FEATURIZATION)
            )
